
(* Heuristique : pour aller vers le pire cas, on va simplement tester tous les coups
   possibles (ou presque) et choisir le scénario qui maximise le temps de stabilisation. *)

open RdbgRun
open RdbgStdLib

type node_info = {
  lvl: int;
  s: bool;
  r: bool
}

let computed_moves: (string * string) list option ref = ref None

let daemon_choice = ref ("", "")

let set_daemon_choice (node: string) (state: string) =
  daemon_choice := node, state

(* Liste des activables sous forme de couples (noeud,état) *)
let enabled (e: RdbgEvent.t): (string * string) list =
  List.filter_map (fun (n,v) ->
    match v with
    | Data.B true -> (
        match String.split_on_char '_' n with
        | "Enab"::pid::state::[] -> Some (pid, state)
        | _ -> None
      )
    | _ -> None
  ) e.data

(* Heuristique qui élimine des coups lors de l'exploration, pour en réduire le coût. 
Comme la fonction de potentiel ne mène pas toujours au pire cas sur le Toy Exemple,
on se base sur les noeuds activables en essayant de faire l'action de propagation "R"
en priorité. On pourrait aussi changer la fonction de potentiel et l'utiliser ici, ce
qui rendrait la méthode beaucoup plus générale. *)
let exploration_hint (enabled_list: (string * string) list) =
  let l = List.filter (fun (_, state) -> state = "Rp" || state = "Rr") enabled_list in
  if l = [] then enabled_list else l


(* Explore un coup, donne le nombre de coups total qu'il engendre (pour aller jusqu'à la fin) *)
let rec explore_count e i moves: int =
  let possible_choices = enabled e in
  match possible_choices with
  | [] ->
    moves (* fini *)
  | l ->
    e.save_state i;
    (* tout explorer coûte trop cher, on élimine les coups les moins intéressants *)
    let filtered = exploration_hint l in
    
    (* calcule le score de tous les coups sélectionnés *)
    let score = List.map (fun (node, st) ->
      set_daemon_choice node st;
      let next = next_cond e (fun x -> x.lang = "sasa") in
      let to_end = explore_count next (i+1) (moves+1) in
      e.restore_state i;
      to_end
    ) filtered in

    (* sélectionne le pire cas *)
    match score with
    | [] -> assert false
    | x::xs ->
      let worst_score = List.fold_left max x xs in
      worst_score

 
(* Table des états déjà explorés, pour éviter de recalculer pour rien.
On stocke les data des outputs en clé et les coups suivants en valeur. *)
let explored = Hashtbl.create 64

(* Donne la partie "outputs" de e.data *)
let event_outputs (e: RdbgEvent.t) =
  (* On utilise le fait que les inputs viennent toujours en premier, puis les outputs *)
  let ins = (List.length e.inputs) + 3 in (* +3 pour enlever silent,legitimate,potential *)
  (List.filteri (fun i _ -> i >= ins) e.data)
  
  (* Autre méthode : filtrer par nom de variable et ne prendre que l'état des noeuds *)
  (* let states = ["input"; "sub"; "res"] in
  let res = List.filter_map (fun (n,v) ->
    let s = String.split_on_char '_' n in
    match s with
    | _::x::[] when List.mem x states -> Some (x,v)
    | _ -> None
  ) e.data
  in
  res *)
 

(* Explore un état, donne le pire cas qu'il engendre
(liste tous les coups suivants menant au pire cas possible à partir de cette configuration) *)
let rec explore e i: (string * string) list =
  (* si on a déjà été dans la même situation avant, on réutilise le résultat *)
  let outputs = event_outputs e in
  match Hashtbl.find_opt explored outputs with
  | Some x -> x
  | None ->
    (
    let possible_choices = enabled e in
    match possible_choices with
    | [] -> [] (* fini: rien ensuite *)
    | l ->
      e.save_state i;
      (* tout explorer coûte trop cher, on élimine les coups les moins intéressants *)
      let filtered = exploration_hint l in
      
      (* calcule les coups suivants pour tous les coups sélectionnés *)
      let all_moves = List.map (fun (node, st) ->
        set_daemon_choice node st;
        let next = next_cond e (fun x -> x.lang = "sasa") in
        (* calcule jusqu'à la fin et choisit la pire suite de coups à partir d'ici *)
        let to_end = explore next (i+1) in
        (* restore l'évènement originel *)
        e.restore_state i;
        (* renvoie ce coup-là avec la pire suite, dans l'ordre chronologique (prepend pour + de perf) *)
        let elem = (node,st) in
        elem::to_end
      ) filtered in

      (* sélectionne le pire cas *)
      match all_moves with
      | [] -> assert false
      | x::xs ->
        let compare a b = 
          let na = List.length a in
          let nb = List.length b in
          if nb > na then b else a
        in
        let worst_case = List.fold_left compare x xs in
        (* stocke le résultat pour éviter de le recalculer si on retombe sur cette config *)
        Hashtbl.add explored outputs worst_case;
        worst_case
    )


(* Entrée : le contenu de !e.data
   Sortie : tuples ("n_S", activate) avec n le nom du noeud et S le nom de l'état, activate un booléen
   Attention, les noms ne doivent contenir que des lettres à cause du Data.type_to_string dans rdbgRun.ml:37 *)
let custom_daemon (data: Data.subst list): Data.subst list option =
  (* liste des noeuds-états *)
  let sl = List.filter (fun (n,v) -> String.length n > 5 && String.sub n 0 5 = "Enab_") data in

  (* évènement sasa courant *)
  let evt = !e in

  (* sortie du démon : active le noeud sélectionné *)
  let get_daemon_choice : string * string =
    match !computed_moves with
    | None ->
      (* lance l'exploration si elle n'est pas déjà en cours *)
      let a,b = !daemon_choice in
      if a = "" && b = "" then (
        Printf.printf "computing worst case...\n";
        let worst_case = explore evt 0 in
        let moves = worst_case in
        Printf.printf "found worst case: %d moves\n" (List.length moves);
        let first_move = List.hd moves in
        computed_moves := Some (List.tl moves);
        first_move
      )
      else
        a,b
    | Some (next_move::rest) ->
      computed_moves := Some rest;
      next_move
    | Some [] ->
      Printf.printf "no event left\n";
      assert false
  in
  
  let sel_pid, sel_state = get_daemon_choice in
  let res = List.map (fun (n,v) ->
    match String.split_on_char '_' n with
    | _::pid::state::[] ->
      let str = Printf.sprintf "%s_%s" pid state in 
      let activate = (pid = sel_pid) && (sel_state = state) in
      (str, Data.B activate)
    | _ -> assert false
  ) sl in

  Some res

let is_end (e: RdbgEvent.t) =
  match (List.assoc_opt "potential" e.data) with
  | Some (Data.F 0.0) -> true
  | _ -> false

let auto () =
  let stop = ref false in
  while not !stop do
    e := next_cond !e (fun e -> e.lang = "sasa");
    pe ();
    stop := is_end !e;
  done;
  Printf.printf "done in %d moves\n" (!e.step - 1);
  ()

let _ =
  rdbg_mv_hook := Some custom_daemon;
