(* [algo_to_deco dir] associates a directory names to 
   - an argument to provide to gg-deco (to decorate gg generated graphs)
   - a label that is used to create the dot file names 
 *)
let algo_to_deco dir =
  match Filename.basename dir with
  | "bfs-spanning-tree" 
  | "dijkstra-ring"     
  | "bfs-st-HC92"       
  | "st-CYH91"          
  | "st-KK06-algo1"     
  | "st-KK06-algo2"     
  | "dfs"               
  | "dfs-list" -> "1-:p.ml 0:root.ml", "p_root"
  | "async-unison" 
  | "coloring"     
  | "alea-coloring"
  | "alea-coloring-unif" -> "0-:p.ml","p"
  | "alea-coloring-alt"  -> "0-:algo_331.ml","algo_331"
  | "unison"        -> "0-:unison.ml","unison"
  | s  -> Printf.printf
            "Error: %s is not handled by GenExpeMakefiles.algo_to_deco\n%!" s;
    assert false

type dir = string
type daemon = string
type graph =
  | Udg of int  (* size of the graph *)
  | Qudg of int (* size of the graph *)
  | ER of int * float (* size of the graph, probability to add an edge *)
  | Grid of int * int (* width and height of the grid *)
  | Clique of int  (* size of the graph *)
  | Ring of int    (* size of the graph *)
;;

(* The  4 following simulation  parameters are put into  references so
   that  they can  be modified.   They  could have  been arguments  of
   gen_makefile, but their values ought to be always the same (this is
   arguable).  *)
let precision = ref 0.01
    (* 0.01 here  means that we simulate until the  Confidence Interval
       size of  the 3 complexity  numbers under estimation  is smaller
       than 1% of their current estimation.
    *)
let directed = ref false (* *)
let max_simu_nb = ref 10000 (* no more simulations are done once reached *)
let timeout_in_sec = ref 10000 (* Ditto, once overtaken *)
let regen_dot = ref true;; (* regenerate the graph before every simulation *)
#use "genExpeMakefilesUtils.ml";;

(* The Main function (or in other words, the entry point)

   [gen_makefile oc p dl dirl  gl targets targets_cmxs] writes on [oc]
   |dl| x |dirl| x |gl| Makefile  rules that build .log files. It also
   generates the  dependencies to  execute those  log rules  (to build
   .dot and .cmxs files).

   It returns  the list of log  and cmxs targets that  were written on [oc]

   More   precisely,  it   adds  to   the  input   list  [targets_log]
   [targets_cmxs] the log and cmxs  targets that were written on [oc].
   *)
let (gen_makefile : string -> daemon list -> dir list -> graph list -> unit ) =
  fun fn dl dirl gl ->
  let oc = open_out fn in
  Printf.fprintf oc "# Generated by coloring_campaign.ml";
  let targets_log, targets_cmxs = List.fold_left (gen_makefile0 oc dl dirl) ([], []) gl in
  Printf.fprintf oc "\n# ZZZ do not do 'make -j 20 $(CMXS)'! (p.cmxs is regenerated
# at each target, which causes failures)\n";
  Printf.fprintf oc "\n# But 'make -j 20 $(LOG)' is fine!\n%!";
  Printf.fprintf oc "\nCMXS=%s\n"  (String.concat " " targets_cmxs);
  Printf.fprintf oc "\nLOG=%s\n%!" (String.concat " " targets_log);
  Printf.printf "%s has been generated\n%!" fn;
  close_out oc

(** Ditto, by building the list of deps incrementally. Requires the caller to print the 
CMXS and LOG makefile files 
 *)
let (gen_makefile_incr : out_channel -> daemon list -> dir list
     -> graph list -> string list -> string list -> string list * string list) =
  fun  oc dl dirl gl targets_log targets_cmxs ->
  let targets_log, targets_cmxs  =
    List.fold_left (gen_makefile0 oc dl dirl) (targets_log, targets_cmxs) gl
  in
  Printf.fprintf oc "\n# ZZZ do not do 'make -j 20 $(CMXS)'! (p.cmxs is regenerated
# at each target, which causes failures)\n";
  Printf.fprintf oc "\n# But 'make -j 20 $(LOG)' is fine!\n%!";
  targets_log, targets_cmxs
